import os

from dagster import ExecutionTargetHandle
from dagster_dask import execute_on_dask, DaskConfig


def test_dask_cluster():
    result = execute_on_dask(
        ExecutionTargetHandle.for_pipeline_module(
            'dagster_examples.toys.hammer', 'hammer_pipeline'
        ),
        env_config={'storage': {'s3': {'config': {'s3_bucket': 'dagster-airflow-scratch'}}}},
        dask_config=DaskConfig(address='%s:8786' % os.getenv('DASK_ADDRESS')),
    )
    assert result.success
    assert result.result_for_solid('total').output_value() == 4



from dagster import file_relative_path, pipeline, solid, ExecutionTargetHandle, InputDefinition
from dagster.core.test_utils import nesting_composite_pipeline
from dagster_dask import execute_on_dask, DaskConfig

import dagster_pandas as dagster_pd


@solid
def simple(_):
    return 1


@pipeline
def dask_engine_pipeline():
    return simple()  # pylint: disable=no-value-for-parameter


def test_execute_on_dask():
    result = execute_on_dask(
        ExecutionTargetHandle.for_pipeline_python_file(__file__, 'dask_engine_pipeline'),
        env_config={'storage': {'filesystem': {}}},
        dask_config=DaskConfig(timeout=30),
    )
    assert result.result_for_solid('simple').output_value() == 1


def dask_composite_pipeline():
    return nesting_composite_pipeline(6, 2)


def test_composite_execute():
    result = execute_on_dask(
        ExecutionTargetHandle.for_pipeline_python_file(__file__, 'dask_composite_pipeline'),
        env_config={'storage': {'filesystem': {}}},
        dask_config=DaskConfig(timeout=30),
    )
    assert result.success


@solid(input_defs=[InputDefinition('df', dagster_pd.DataFrame)])
def pandas_solid(_, df):  # pylint: disable=unused-argument
    pass


@pipeline
def pandas_pipeline():
    return pandas_solid()  # pylint: disable=no-value-for-parameter


def test_pandas_dask():
    environment_dict = {
        'solids': {
            'pandas_solid': {
                'inputs': {'df': {'csv': {'path': file_relative_path(__file__, 'ex.csv')}}}
            }
        }
    }

    result = execute_on_dask(
        ExecutionTargetHandle.for_pipeline_python_file(__file__, pandas_pipeline.name),
        env_config={'storage': {'filesystem': {}}, **environment_dict},
        dask_config=DaskConfig(timeout=30),
    )

    assert result.success



import itertools

import dask
import dask.distributed

from dagster import check
from dagster.core.engine.engine_base import IEngine

from dagster.core.events import DagsterEvent
from dagster.core.execution.context.system import SystemPipelineExecutionContext
from dagster.core.execution.plan.plan import ExecutionPlan

from dagster_graphql.client.mutations import execute_start_pipeline_execution_query

from .config import DaskConfig


def query_on_dask_worker(handle, variables, dependencies):  # pylint: disable=unused-argument
    '''Note that we need to pass "dependencies" to ensure Dask sequences futures during task
    scheduling, even though we do not use this argument within the function.

    We also set 'raise_on_error' within pipeline execution, because otherwise (currently) very
    little information is propagated to the dask master from the workers about the state of
    execution; we should at least inform the user of exceptions.
    '''
    return execute_start_pipeline_execution_query(handle, variables)


class DaskEngine(IEngine):  # pylint: disable=no-init
    @staticmethod
    def execute(pipeline_context, execution_plan, step_keys_to_execute=None):
        check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
        check.inst_param(execution_plan, 'execution_plan', ExecutionPlan)
        check.opt_list_param(step_keys_to_execute, 'step_keys_to_execute', of_type=str)

        step_key_set = None if step_keys_to_execute is None else set(step_keys_to_execute)

        dask_config = pipeline_context.run_config.executor_config

        check.param_invariant(
            isinstance(pipeline_context.executor_config, DaskConfig),
            'pipeline_context',
            'Expected executor_config to be DaskConfig got {}'.format(
                pipeline_context.executor_config
            ),
        )

        # Checks to ensure storage is compatible with Dask configuration
        storage = pipeline_context.environment_dict.get('storage')
        check.invariant(storage.keys(), 'Must specify storage to use Dask execution')

        if dask_config.is_remote_execution:
            check.invariant(
                storage.get('s3'),
                'Must use S3 storage with non-local Dask address {dask_address}'.format(
                    dask_address=dask_config.address
                ),
            )
        else:
            check.invariant(
                not storage.get('in_memory'),
                'Cannot use in-memory storage with Dask, use filesystem or S3',
            )

        step_levels = execution_plan.topological_step_levels()

        pipeline_name = pipeline_context.pipeline_def.name

        with dask.distributed.Client(**dask_config.build_dict(pipeline_name)) as client:
            execution_futures = []
            execution_futures_dict = {}

            for step_level in step_levels:
                for step in step_level:
                    if step_key_set and step.key not in step_key_set:
                        continue

                    step_context = pipeline_context.for_step(step)

                    check.invariant(
                        not step_context.run_config.loggers,
                        'Cannot inject loggers via RunConfig with the Dask executor',
                    )

                    check.invariant(
                        not step_context.event_callback,
                        'Cannot use event_callback with Dask executor',
                    )

                    # We ensure correctness in sequencing by letting Dask schedule futures and
                    # awaiting dependencies within each step.
                    dependencies = [
                        execution_futures_dict[step_input.prev_output_handle.step_key]
                        for step_input in step.step_inputs
                        if step_input.prev_output_handle is not None
                    ]

                    variables = {
                        'executionParams': {
                            'selector': {'name': pipeline_name},
                            'environmentConfigData': pipeline_context.environment_dict,
                            'mode': pipeline_context.mode_def.name,
                            'executionMetadata': {'runId': pipeline_context.run_config.run_id},
                            'stepKeys': [step.key],
                        }
                    }

                    dask_task_name = '%s.%s' % (pipeline_name, step.key)

                    future = client.submit(
                        query_on_dask_worker,
                        pipeline_context.execution_target_handle,
                        variables,
                        dependencies,
                        key=dask_task_name,
                    )

                    execution_futures.append(future)
                    execution_futures_dict[step.key] = future

            # This tells Dask to awaits the step executions and retrieve their results to the master
            execution_step_events = client.gather(execution_futures)

            # execution_step_events is now a list of lists, the inner lists contain the dagster
            # events emitted by each step execution
            for step_event in itertools.chain.from_iterable(execution_step_events):
                check.inst(step_event, DagsterEvent)

                yield step_event



from dagster import check, ExecutionTargetHandle, RunConfig, SystemStorageData

from dagster.core.execution.api import create_execution_plan, scoped_pipeline_context
from dagster.core.execution.results import PipelineExecutionResult

from .config import DaskConfig
from .engine import DaskEngine


def execute_on_dask(
    handle, env_config=None, run_config=None, dask_config=None
):  # pylint: disable=too-many-locals
    check.inst_param(handle, 'handle', ExecutionTargetHandle)
    check.opt_dict_param(env_config, 'env_config', key_type=str)
    dask_config = check.opt_inst_param(dask_config, 'dask_config', DaskConfig, DaskConfig())
    run_config = check.opt_inst_param(
        run_config, 'run_config', RunConfig, RunConfig(executor_config=dask_config)
    )

    check.inst(
        run_config.executor_config,
        DaskConfig,
        'run_config.executor_config should be instance of DaskConfig to execute on Dask',
    )

    pipeline_def = handle.build_pipeline_definition()

    execution_plan = create_execution_plan(pipeline_def, env_config, run_config=run_config)

    with scoped_pipeline_context(pipeline_def, env_config, run_config) as pipeline_context:
        events = list(DaskEngine.execute(pipeline_context, execution_plan, None))

        return PipelineExecutionResult(
            pipeline_def,
            run_config.run_id,
            events,
            lambda: scoped_pipeline_context(
                pipeline_def,
                env_config,
                run_config,
                system_storage_data=SystemStorageData(
                    intermediates_manager=pipeline_context.intermediates_manager,
                    run_storage=pipeline_context.run_storage,
                    file_manager=pipeline_context.file_manager,
                ),
            ),
        )
